/*Open assistive arm for meals code
 *  This firmware is used to calibrate the device before using by someone.
 *  version 1 08/09/2020
 *  edited by Georges Tawil 
 *  contact : georges.tawil@apf.asso.fr
 * 
 * 
 * 
 */
// Import libraries:
//    - dynamixel shield library can be get from here : https://emanual.robotis.com/docs/en/parts/interface/dynamixel_shield/
#include <Dynamixel2Arduino.h>

// Define DXL Motors communication
#define       DXL_SERIAL   Serial
const uint8_t DXL_DIR_PIN = 2;              // DYNAMIXEL Shield DIR PIN
const float   DXL_PROTOCOL_VERSION = 2.0; 

typedef struct{
  uint8_t   id                      ;
  bool      availability            ;
  int       max_speed               ;   // Profile Velocity
  float     calibrationParameter[3] ;
  bool      enable                  ;
  float     current_angle           ;
  } Actuator;
  
Dynamixel2Arduino dxl(DXL_SERIAL, DXL_DIR_PIN);

//*********************************Interrupts****************************************************
#define interruptPin 18 // defining pin 18 as an interrupt pin
unsigned long t_high=0, t_low=0, t=0, t_but = 0;
bool but_rise = false,  but_fall = false;

//time intervals of pressing the button (milliseconds)
#define BETWEEN_0_200       0
#define BETWEEN_200_700     1
#define BETWEEN_700_4000    2

int  but_status = BETWEEN_0_200;          // Init button status
bool but_last  = false, but_new = false;

bool action_required = false, action_done = false;
//***********************************************************************************************

#define ALL      0

// Create a table for describing all actuators of the robot arm
Actuator actuator_tab[6];

bool actuator_available=false;

float cmd[6];
float cmd3[3];

// using cycle
float grabing_position[3]={-32.97862275, 0.74171066, 60.84672491};

float grabing_position2[3]={-32.97862275, 10.83920747, 60.91914321};

int grabing_position_counter=0;
float transition_position[3]={0.0, 0.0, 0.0};
float home_position[3]={0.0, 0.0, 0.0};
bool finish=false;

//saved position
float mouth_position[6];
float mouth_position2[3];
bool mouth_position_saved = false;


void setup() {
  // Set Port baudrate to 57600bps. This has to match with DYNAMIXEL baudrate.
  dxl.begin(57600);
  // Set Port Protocol Version. This has to match with DYNAMIXEL protocol version.
  dxl.setPortProtocolVersion(DXL_PROTOCOL_VERSION);

  // Interrupts Init
  pinMode(interruptPin, INPUT);  // pin definition as input
  attachInterrupt(digitalPinToInterrupt(interruptPin), button_change, CHANGE);  // call function button_change() in the state of button was changed

  
  //                {id , availability  , max_speed  ,  calibrationParameter  ,  enable , current_angle };
  actuator_tab[0] = {1  ,     false     ,     35     ,    {3.64, 170.72, 0.0}     ,   false ,        0     };
  actuator_tab[1] = {2  ,     false     ,     35     ,    {2.06, 0.0, 247.81}     ,   false ,        0     };
  actuator_tab[2] = {3  ,     false     ,     35     ,    {-3.96, -2.07, 408.32}     ,   false ,        0     };
  actuator_tab[3] = {4  ,     false     ,     35     ,    {1.0, 181.9, 0.0}     ,   false ,        0     };
  actuator_tab[4] = {5  ,     false     ,     35     ,    {1.0, 180.58, 0.0}     ,   false ,        0     };
  actuator_tab[5] = {6  ,     false     ,     35     ,    {1.0, 151.96, 0.0}     ,   false ,        0     };

  actuator_available=actuator_availability_checking();
}

void loop() {
  // verification que les actionneurs sont opérationnels et n'ont pas de défaillance
  if (actuator_available){
    if (finish==false){
      // cycle de service 
      if (mouth_position_saved){
      chargement();
      service();
      check_times();
      }
      // enregistrement de la position de bouche
      else{
        read_data_motor();
        check_times();
        if (action_required && !action_done){
          do_action();
          action_done = true;
        }
      }
    }
  }
  else{
    while(1){}
  }
}

void chargement(){
  float angle=45.0;
  float commande;
  moveTo3(grabing_position);
  read_data_motor();
  commande=(-angle+(actuator_tab[2].calibrationParameter[2]-actuator_tab[2].current_angle)/actuator_tab[2].calibrationParameter[0]+((actuator_tab[1].current_angle-actuator_tab[1].calibrationParameter[2])/actuator_tab[1].calibrationParameter[0])*((actuator_tab[2].calibrationParameter[1]/actuator_tab[2].calibrationParameter[1])-1))*actuator_tab[5].calibrationParameter[0]+actuator_tab[5].calibrationParameter[1];
  dxl.setGoalPosition(actuator_tab[5].id, commande, UNIT_DEGREE);
  moveTo5(grabing_position2);
  read_data_motor();
  commande=((actuator_tab[2].calibrationParameter[2]-actuator_tab[2].current_angle)/actuator_tab[2].calibrationParameter[0]+((actuator_tab[1].current_angle-actuator_tab[1].calibrationParameter[2])/actuator_tab[1].calibrationParameter[0])*((actuator_tab[2].calibrationParameter[1]/actuator_tab[2].calibrationParameter[1])-1))*actuator_tab[5].calibrationParameter[0]+actuator_tab[5].calibrationParameter[1];
  dxl.setGoalPosition(actuator_tab[5].id, commande, UNIT_DEGREE);
  delay(500);
  moveTo3(grabing_position);
  finish=true;
}

void service(){
  moveTo3(transition_position);
  moveTo4(mouth_position2);
  delay(2000);
  moveTo3(transition_position);
}

/* Actuator availability checking
 *  This function checks that every actuators are wiring and available to be controled, it returns an boolean:
 *    -false : there is one or more actuator is are not available
 *    -true : all actuators are available, they can be used
 *  la fonction permet de vérifier que tous les actionneurs sont connectés et disponible pour le controle
 */
bool actuator_availability_checking(){
  bool availability = true;
  for (int i=0; i<6; i++){
    actuator_tab[i].availability = dxl.ping(actuator_tab[i].id);
    if (actuator_tab[i].availability){
      dxl.torqueOff(actuator_tab[i].id);
      dxl.setOperatingMode(actuator_tab[i].id, OP_EXTENDED_POSITION);
      dxl.writeControlTableItem(PROFILE_VELOCITY,     actuator_tab[i].id, actuator_tab[i].max_speed);
      dxl.writeControlTableItem(PROFILE_ACCELERATION, actuator_tab[i].id, actuator_tab[i].max_speed/2);
      dxl.torqueOn(actuator_tab[i].id);
    }
    else{
      availability = false;
    }
  }
  if (availability){
    Serial.println();
    Serial.println("***************      Actuators are connected and work     ********************");
    Serial.println();
  }
  return availability;
}

/* Read actuator datas
 *  Read every data of actuators and push them in actuators table with updated data
 */
void read_data_motor(){
  for (int i=0; i<6; i++){
    actuator_tab[i].current_angle  = dxl.getPresentPosition(actuator_tab[i].id, UNIT_DEGREE);   // Read angle 
  }
}

void print_data_motor(){
  Serial.println();
  Serial.print("Present Position:");
  for (int i=0; i<6; i++){
    Serial.print("  q");  Serial.print(actuator_tab[i].id); Serial.print(": "); Serial.print(actuator_tab[i].current_angle);
  }
  Serial.println();
}

void torque_on(int id){
  switch (id){
    case ALL:
      for (int i=0; i<6; i++){
        if (actuator_tab[i].availability){
          dxl.torqueOn(actuator_tab[i].id);
        }
      }
      break;
    default:
      dxl.torqueOn(actuator_tab[id-1].id);
  }
}

void torque_off(int id){
  switch (id){
    case ALL:
      for (int i=0; i<6; i++){
        if (actuator_tab[i].availability){
          dxl.torqueOff(actuator_tab[i].id);
        }
      }
      break;
    default:
      dxl.torqueOff(actuator_tab[id-1].id);
  }
}

void button_change(){
  bool but_new = digitalRead(interruptPin);
  //Serial.print("but_last = "); Serial.print(but_last); Serial.print(" but_new = "); Serial.println(but_new);
  if (but_new > but_last) button_rise();
  else if (but_new < but_last) button_fall();
  but_last = but_new;
}

void button_rise(){
  but_rise   = true;
  but_fall   = false;
  t = millis();
}

void button_fall(){
  but_fall   = true;
  but_rise   = false;
  t = millis();
}

void check_times(){
  long int dt;
  dt = millis() - t;
  if(but_rise){
    action_required = false;
    t_high = dt;
    t_low  = 0;
  }
  else if(but_fall){
    t_low  = dt;
    t_high = 0;
  }

  //Serial.println(dt);

  if      (t_high > 700)                    { torque_off(ALL); }
  
  if      (t_high > 0   && t_high < 200  )  { but_status = BETWEEN_0_200;    }
  else if (t_high > 200 && t_high < 1000  )  { but_status = BETWEEN_200_700; action_done = false; }
  else if (t_high > 2000 && t_high < 4000 )  { but_status = BETWEEN_700_4000; action_done = false;}
  else if (t_high > 4000 ) { but_status = BETWEEN_0_200; }
  if      (t_low > 200 && !action_done)     { torque_on(ALL); action_required = true;}
}

void print_saved_position(int joint_id, float angle){
  Serial.println();
  Serial.println("***************      Position saved     ********************");
  Serial.print("            Joint :   "); Serial.print(joint_id); Serial.print(",        readed position :   ");Serial.println(angle);
  Serial.println();
}

void do_action(){
  switch(but_status){
    case BETWEEN_200_700:
      for (int i=0; i<3; i++){
        mouth_position2[i]=actuator_tab[i].current_angle;
      }
      Serial.println();
      Serial.println("***************      position bouche enregistré     ********************");
      Serial.println();
      
      mouth_position_saved=true;
      delay(5000);
      moveTo3(home_position);
    break;
      
    case BETWEEN_700_4000:
    Serial.println();
    Serial.println("***************      Reprendre la position de bouche     ********************");
    Serial.println();
    mouth_position_saved=false;
    break;
  }
}

void read_joint_position(float *joint_position){
  read_data_motor();
  joint_position[0]=(actuator_tab[0].current_angle-actuator_tab[0].calibrationParameter[1])/actuator_tab[0].calibrationParameter[0];
  joint_position[1]=(actuator_tab[1].current_angle-actuator_tab[1].calibrationParameter[2])/actuator_tab[1].calibrationParameter[0];
  joint_position[2]=(actuator_tab[2].current_angle-actuator_tab[2].calibrationParameter[1]*joint_position[1]-actuator_tab[2].calibrationParameter[2])/actuator_tab[2].calibrationParameter[0];
  joint_position[3]=(actuator_tab[3].current_angle-actuator_tab[3].calibrationParameter[1])/actuator_tab[3].calibrationParameter[0];
  joint_position[4]=(actuator_tab[4].current_angle-actuator_tab[4].calibrationParameter[1])/actuator_tab[4].calibrationParameter[0];
  joint_position[5]=(actuator_tab[5].current_angle-actuator_tab[5].calibrationParameter[1])/actuator_tab[5].calibrationParameter[0];
}



void modelPosition2command3(float modelPosition[3]){
  cmd3[0]=modelPosition[0]*actuator_tab[0].calibrationParameter[0]+actuator_tab[0].calibrationParameter[1];
  cmd3[1]=modelPosition[1]*actuator_tab[1].calibrationParameter[0]+modelPosition[2]*actuator_tab[1].calibrationParameter[1]+actuator_tab[1].calibrationParameter[2];
  cmd3[2]=modelPosition[2]*actuator_tab[2].calibrationParameter[0]+modelPosition[1]*actuator_tab[2].calibrationParameter[1]+actuator_tab[2].calibrationParameter[2];
}



float compare_target_current2(){
  read_data_motor();
  float difference=0.0;
  for (int i=0; i<3; i++){
    difference=difference+abs(cmd3[i]-actuator_tab[i].current_angle);
  }
  return difference;
}

void stabiliser2(){
  float commande[3];
  float joint1;
  float joint23;
  read_data_motor();
  joint1=(actuator_tab[0].current_angle-actuator_tab[0].calibrationParameter[1])/actuator_tab[0].calibrationParameter[0];
  joint23=(actuator_tab[2].calibrationParameter[2]-actuator_tab[2].current_angle)/actuator_tab[2].calibrationParameter[0]+((actuator_tab[1].current_angle-actuator_tab[1].calibrationParameter[2])/actuator_tab[1].calibrationParameter[0])*((actuator_tab[2].calibrationParameter[1]/actuator_tab[2].calibrationParameter[1])-1);
  commande[0]=joint23*actuator_tab[3].calibrationParameter[0]+actuator_tab[3].calibrationParameter[1];
  commande[1]=-joint1*actuator_tab[4].calibrationParameter[0]+actuator_tab[4].calibrationParameter[1];
  commande[2]=actuator_tab[5].calibrationParameter[1];
  for (int i=0; i<3; i++){
    dxl.setGoalPosition(actuator_tab[i+3].id, commande[i], UNIT_DEGREE);
  }
}

void stabiliser1(){
  float commande[3];
  float joint23;
  read_data_motor();
  joint23=(actuator_tab[2].calibrationParameter[2]-actuator_tab[2].current_angle)/actuator_tab[2].calibrationParameter[0]+((actuator_tab[1].current_angle-actuator_tab[1].calibrationParameter[2])/actuator_tab[1].calibrationParameter[0])*((actuator_tab[2].calibrationParameter[1]/actuator_tab[2].calibrationParameter[1])-1);
  commande[0]=actuator_tab[3].calibrationParameter[1];
  commande[1]=actuator_tab[4].calibrationParameter[1];
  commande[2]=joint23*actuator_tab[5].calibrationParameter[0]+actuator_tab[5].calibrationParameter[1];
  for (int i=0; i<3; i++){
    dxl.setGoalPosition(actuator_tab[i+3].id, commande[i], UNIT_DEGREE);
  }
}

void moveTo3(float Q[3]){
  float epsilon=10.0;
  float ecart;
  modelPosition2command3(Q);
  for (int i=0; i<3; i++){
    dxl.setGoalPosition(actuator_tab[i].id, cmd3[i], UNIT_DEGREE);
  }
  stabiliser1();
  ecart=compare_target_current2();
  while (ecart>epsilon){
    for (int i=0; i<3; i++){
      dxl.setGoalPosition(actuator_tab[i].id, cmd3[i], UNIT_DEGREE);
    }
    stabiliser1();
    ecart=compare_target_current2();
    Serial.println(ecart);
  }
}

void moveTo5(float Q[3]){
  float epsilon=20.0;
  float ecart;
  modelPosition2command3(Q);
  for (int i=0; i<3; i++){
    dxl.setGoalPosition(actuator_tab[i].id, cmd3[i], UNIT_DEGREE);
  }
  ecart=compare_target_current2();
  while (ecart>epsilon){
    for (int i=0; i<3; i++){
      dxl.setGoalPosition(actuator_tab[i].id, cmd3[i], UNIT_DEGREE);
    }
    ecart=compare_target_current2();
    Serial.println();
    Serial.println(ecart);
  }
}

void moveTo4(float Q[3]){
  float epsilon=10.0;
  float ecart;
  for (int i=0; i<3; i++){
    cmd3[i]=Q[i];
    dxl.setGoalPosition(actuator_tab[i].id, cmd3[i], UNIT_DEGREE);
  } 
  stabiliser2();
  ecart=compare_target_current2();
  while (ecart>epsilon){
    for (int i=0; i<3; i++){
      cmd3[i]=Q[i];
      dxl.setGoalPosition(actuator_tab[i].id, cmd3[i], UNIT_DEGREE);
    } 
    stabiliser2();
    ecart=compare_target_current2();
    Serial.println(ecart);
  }   
}



/*
void moveTo(float Q[6]){
  float epsilon=10.0;
  float ecart;
  modelPosition2command(Q);
  for (int i=0; i<6; i++){
    dxl.setGoalPosition(actuator_tab[i].id, cmd[i], UNIT_DEGREE);
  }
  ecart=compare_target_current();
  while (ecart>epsilon){
    ecart=compare_target_current();
    Serial.println(ecart);
  }
}

float compare_target_current(){
  float difference=0.0;
  for (int i=0; i<6; i++){
    difference=difference+abs(cmd[i]-actuator_tab[i].current_angle);
  }
  return difference;
}

void modelPosition2command(float modelPosition[6]){
  cmd[0]=modelPosition[0]*actuator_tab[0].calibrationParameter[0]+actuator_tab[0].calibrationParameter[1];
  cmd[1]=modelPosition[1]*actuator_tab[1].calibrationParameter[0]+modelPosition[2]*actuator_tab[1].calibrationParameter[1]+actuator_tab[1].calibrationParameter[2];
  cmd[2]=modelPosition[2]*actuator_tab[2].calibrationParameter[0]+modelPosition[1]*actuator_tab[2].calibrationParameter[1]+actuator_tab[2].calibrationParameter[2];
  cmd[3]=modelPosition[3]*actuator_tab[3].calibrationParameter[0]+actuator_tab[3].calibrationParameter[1];
  cmd[4]=modelPosition[4]*actuator_tab[4].calibrationParameter[0]+actuator_tab[4].calibrationParameter[1];
  cmd[5]=modelPosition[5]*actuator_tab[5].calibrationParameter[0]+actuator_tab[5].calibrationParameter[1];
}

void moveTo2(float Q[6]){ 
  for (int i=0; i<6; i++){
    dxl.setGoalPosition(actuator_tab[i].id, Q[i], UNIT_DEGREE);
  }
}

*/


void printMessage(const char *message){
  Serial.println();
  Serial.print("************************         ");
  Serial.print(message);
  Serial.println("         ************************");     
  Serial.println();
}

void print_position(const char *message, float pos[6]){
  printMessage(message);
  for (int i=0; i<6; i++){
    Serial.print("          Joint ");
    Serial.print(actuator_tab[i].id);
    Serial.print("  :");
    Serial.println(pos[i]);
  }
}
